/*  дерево символов, слов и фраз
вербальная иерархия распознавателей

*/

package word_sensor

import (
	"BOT/lib"
	"strconv"
	"strings"
)

/////////////////////////////////////////////
// Первый уровень дерева - символы. Заглавных не может быть в воспринимаемом
var symbolsArr = []string{" ","а","б","в","г","д","е","ё","ж","з","и","й","к","л","м","н","о","п","р","с","т","у","ф","х","ц","ч","ш","щ","ъ","ы","ь","э","ю","я","a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","w","x","y","z","!","?","@","#","$","%","^","&","*","(",")","+","=","-","1","2","3","4","5","6","7","8","9","0","[","]","{","}","<",">",".",",","/"}
func IsSymbol(w string)(bool){
	for i := 0; i < len(symbolsArr); i++ {
		if w == symbolsArr[i]{
			return true
		}
	}
	return false
}
/////////////////////////////////////////




////////////////////////////////////////////////////////////////////
type WordTree struct {
	ID int // id слова-фразы
	Word string
	Children []WordTree // дочерние узлы (ветвление) НЕ АДРЕСА, А РЕАЛЬНЫЕ ОБЪЕКТЫ
	ParentID int     // ID родителя
	ParentNode *WordTree  // адрес родителя
}

var VernikeWordTree WordTree
var WordTreeFromID=make(map[int]*WordTree)
var WordTreeFromStr=make(map[string][]*WordTree)// массив узлов с такой частью слова
/////////////// для обеспечения уникальности узлов:
type WordUnicum struct {
	parentID int
	word string
}
var WordUnicumIdStr=make(map[WordUnicum]int)// для каждого сочетания  выдается ID узла
//////////////////////////////////////////////////
var lastIDwordTree=0
///////////////////////////////////////////////////////////////////////

// подошла очередь инициализации
func afterLoadTempArr(){
	loadWordTree()
/*
	//а я нехочу
	new:=createNewNodeWordTree(WordTreeFromID[2],0," я")
	new=createNewNodeWordTree(new,0," нехочу")
	createNewNodeWordTree(new,0," нечего знать")
	saveWordTree()


	last:=SetNewWordTreeNode("их") // а я нехочу такое дело
	saveWordTree()
if last!=nil{}
	return
	*/

 //   updateWordTreeFromTempArr(2,4)

afetrInitWordTree()
}

// создать первый, символьный уровень дерева
func initWordTree(vt *WordTree){
	for i := 0; i < len(symbolsArr); i++ {
		createNewNodeWordTree(vt,0,symbolsArr[i])
	}

	return
}
/////////////////////////////////

func createNewNodeWordTree(parent *WordTree,id int,word string)(*WordTree){
	if parent==nil{
		return nil
	}
	// не позволять дублировать узлы
	unic:=WordUnicumIdStr[WordUnicum{parent.ID,word}]
	if unic!=0 {// уже есть такой узел
		old:=WordTreeFromID[unic]
		return old //nil
	}

	if id==0{
		lastIDwordTree++;
		id=lastIDwordTree
	}else{
//		newW.ID=id
		if lastIDwordTree<id{
			lastIDwordTree=id
		}
	}
	var newW WordTree
		newW.ID = id
		newW.ParentID = parent.ID
		newW.ParentNode = parent
		newW.Word = word

		parent.Children = append(parent.Children, newW)
		// четко находим новый вставленный член (а не &parent.Children[count-1])
		var new *WordTree
		for i := 0; i < len(parent.Children); i++ {
			if parent.Children[i].ID == newW.ID {
				new = &parent.Children[i]
			}
		}

//WordTreeFromID[new.ID]=new
	// т.к. append меняет длину массива, перетусовывая адреса, то нужно:
	updateWordTreeFromID()// здесь потому, что при загрузке из файла нужно на лету получать адреса

	return new
}
// корректируем адреса всех узлов
func updateWordTreeFromID(){
	updatingWordTreeFromID(&VernikeWordTree)
}
// проход всего дерева
func updatingWordTreeFromID(wt *WordTree){
	if wt.ID>0 {
		WordTreeFromID[wt.ID] = wt
		WordTreeFromStr[wt.Word] = append(WordTreeFromStr[wt.Word],wt)
		WordUnicumIdStr[WordUnicum{wt.ParentID,wt.Word}] = wt.ID
	}
	if wt.Children == nil{// конец ветки
		return
	}
	for i := 0; i < len(wt.Children); i++ {
		updatingWordTreeFromID(&wt.Children[i])
	}
}
///////////////////////////////////////////////////////////


func loadWordTree(){

	initWordTree(&VernikeWordTree)
	strArr,_:=lib.ReadLines(lib.GetMainPathExeFile()+"/memory_reflex/word_tree.txt")
	cunt:=len(strArr)
//просто проход по всем строкам файла подряд так что сначала идут дочки, потом - их родители
	for n := 0; n < cunt; n++ {
		if len(strArr[n])<2{
			return
		}
		p:=strings.Split(strArr[n], "|#|")
		word:=p[1]
		idP:=strings.Split(p[0], "|")
		id,_:=strconv.Atoi(idP[0])
		parentID,_:=strconv.Atoi(idP[1])
		// новый узел с каждой строкой из файла
		createNewNodeWordTree(WordTreeFromID[parentID],id,word)
	}
	return
}
////////////////////////////////////////////
func saveWordTree(){
	var out=""
// записываем, начиная со второго уровня (уровень символов не пишем)
cnt:=len(VernikeWordTree.Children)
	for n := 0; n < cnt; n++ {
		out+=getWtreeNode(&VernikeWordTree.Children[n])
	}
	lib.WriteFileContent(lib.GetMainPathExeFile()+"/memory_reflex/word_tree.txt",out)
	return
}
func getWtreeNode(wt *WordTree)(string){
	var out=""
	if wt.ParentID>0 {
		out += strconv.Itoa(wt.ID) + "|"
		out += strconv.Itoa(wt.ParentID) + "|#|"
		out += wt.Word + "\r\n"
	}
	if(wt.Children==nil){// конец
		return out
	}
	for n := 0; n < len(wt.Children); n++ {
		out+=getWtreeNode(&wt.Children[n])
	}
return out
}
//////////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////////

/* переносим в дерево слов достаточно повторяющиеся из tempArr
 limitWord - число повторений отдельных слов, с которых начинается пеернос.
limitFraze - число повторений отдельных фраз (с несколькими словами), с которых начинается пеернос.
Для автозаливки - updateWordTreeFromTempArr(4,6),
для сообщения с Пульта - updateWordTreeFromTempArr(2,4)
 */
func updateWordTreeFromTempArr(limitWord int,limitFraze int){
// сначала прходим только слова, потом - только фразы,
// чтобы фразы могли использовать слова (минимизация размера дерева)
	for k, v := range tempArr {
// слово или фраза?
		sps:=strings.Split(k, " ")
		if len(sps)==0{// это слово
			if v >= limitFraze {
				SetNewWordTreeNode(k)
			}
		}
	}

	for k, v := range tempArr {
		// слово или фраза?
		sps:=strings.Split(k, " ")
		if len(sps)>0{// это фраза
			if v >= limitFraze {
				SetNewWordTreeNode(k)
			}
		}
	}
	saveWordTree()
}
/////////////////////////////////////////////////

// вставить новое слово-фразу в дерево:
//найти подходящий узел и если еще нет - вставить новый.
func SetNewWordTreeNode(word string)(*WordTree){
	if len(word)==0{
		return nil
	}
	var last *WordTree
	// первый символ
	r := []rune(word)
	// основа дерева
	cnt:=len(VernikeWordTree.Children)
	for n := 0; n < cnt; n++ {
		rt := []rune(VernikeWordTree.Children[n].Word)
		if r[0]==rt[0]{
			last=findWordTreeNode(r[1:],&VernikeWordTree.Children[n])
		}
	}
	return last
}
func findWordTreeNode(word []rune,wt *WordTree)(*WordTree){
	if len(word)==0{// конец, слово распознано
		return wt
	}
	if(wt.Children==nil ){// конец
		if len(word)>0{// но еще есть фраза -
			// дорастить в дочке узла wt
			createNewNodeWordTree(wt,0,string(word))
			return wt
		}
		return wt // найден узел
	}
	cnt:=len(wt.Children)
	for n := 0; n < cnt; n++ {
		rt := []rune(wt.Children[n].Word)
		if lib.RuneCompare(rt,word)==1{// rt > word - закончить,
			//в этом узле создать новую дочку
			createNewNodeWordTree(wt.Children[n].ParentNode,0,string(word))
			return wt.Children[n].ParentNode
		}
		for i := 0; i < len(rt); i++ {
			rOst:=word[i:len(rt)]
			//if string(rOst)==wt.Children[n].Word{// часть слова совпала со словом в узле
			if lib.RuneCompare(rOst,rt)==0{// часть слова совпала со словом в узле
				// тогда идем дальше по узлам, пройдя совпавшую чать слова
				ost:=word[len(rt):]
				last:=findWordTreeNode(ost,&wt.Children[n])
				return last
			}
		}
	}
	return nil
}
////////////////////////////////////////////////////////////


/*
// вставить новое слово-фразу в дерево:
//найти подходящий узел и если еще нет - вставить новый.
func SetNewWordTreeNode(word string)(*WordTree){
	if len(word)==0{
		return nil
	}
	var last *WordTree
	// первый символ
	r := []rune(word)
	// основа дерева
	cnt:=len(VernikeWordTree.Children)
	for n := 0; n < cnt; n++ {
		rt := []rune(VernikeWordTree.Children[n].Word)
		if r[0]==rt[0]{
			last=findWordTreeNode(r[1:],&VernikeWordTree.Children[n])
		}
	}
	return last
}
func findWordTreeNode(word []rune,wt *WordTree)(*WordTree){
	if len(word)==0{// конец, слово распознано
		return wt
	}
	if(wt.Children==nil ){// конец
		if len(word)>0{// но еще есть фраза -
//НЕ создавать новый узел в середине слова, а делать новый узел предыдущего узла
		// это обрезанное слово?
			//var isCuted=false


			//if isCuted {// дорастить в дочке предыдущего узла wt
			//	createNewNodeWordTree(wt0, 0, string(word))
			//}else{// дорастить в дочке узла wt
				createNewNodeWordTree(wt, 0, string(word))
			//}
			return wt
		}
		return wt // найден узел
	}
	cnt:=len(wt.Children)
	for n := 0; n < cnt; n++ {
		rt := []rune(wt.Children[n].Word)
		res:=lib.RuneCompare(word,rt)
		if res==0{//точное совпадение - закончили проход
			return wt
		}
		if res==1{// фраза больше узла rt
//в этом узле создать новую дочку если помещается целое слово иначе создать новую дочку - в родителе
			//var cutN=0
			for i := 0; i < len(word); i++ {
				//cutN=i
				if i>0 && string(rt[i])==" "{
					w:=string(rt[:i])
					createNewNodeWordTree(&wt.Children[n],0,w)
					return wt.Children[n].ParentNode
					//last:=findWordTreeNode(rt[i:],new)
					//return last
				}
			}
		}
		if res==2 { // фраза меньше узла
			new:=createNewNodeWordTree(wt.Children[n].ParentNode,0,string(word))
			return new
		}

		for i := 0; i < len(rt); i++ {
			rOst:=word[i:len(rt)]
			//if string(rOst)==wt.Children[n].Word{// часть слова совпала со словом в узле
			if lib.RuneCompare(rOst,rt)==0{// часть слова совпала со словом в узле
				// тогда идем дальше по узлам, пройдя совпавшую чать слова
				ost:=word[len(rt):]
				last:=findWordTreeNode(ost,&wt.Children[n])
					return last
			}
		}

	}
	return nil
}
////////////////////////////////////////////////////////////
*/